/*
Copyright 2007-2024. Algoryx Simulation AB.

All AGX source code, intellectual property, documentation, sample code,
tutorials, scene files and technical white papers, are copyrighted, proprietary
and confidential material of Algoryx Simulation AB. You may not download, read,
store, distribute, publish, copy or otherwise disseminate, use or expose this
material unless having a written signed agreement with Algoryx Simulation AB, or having been
advised so by Algoryx Simulation AB for a time limited evaluation, or having purchased a
valid commercial license from Algoryx Simulation AB.

Algoryx Simulation AB disclaims all responsibilities for loss or damage caused
from using this software, unless otherwise stated in written agreements with
Algoryx Simulation AB.
*/

#pragma once

#include <agx/ElementaryConstraint.h>
#include <agx/GenericConstraintData.h>

namespace agx
{
  /**
  Generic elementary constraint with minimal amount of user callbacks
  to reduce development/prototype time and trivially exported using SWIG
  to e.g., Python or C#. This elementary constraint is only valid for
  one or two body constraints.

  Python example for elementary Dot1:
    class MyDot1(agx.GenericElementaryConstraint):
      def __init__(self, dir1, dir2):
          super().__init__(1)
          self.dir1 = dir1
          self.dir2 = dir2

      def onWriteData(self, constraintData: agx.GenericConstraintData):
          a1: agx.Attachment = self.getAttachmentPair().getAttachment1()
          a2: agx.Attachment = self.getAttachmentPair().getAttachment2()
          v1: agx.Vec3 = a1.get(self.dir1)
          v2: agx.Vec3 = a2.get(self.dir2)

          rb1_data: agx.GenericConstraintBodyData = constraintData.getBodyDataAt(index=0)
          rb2_data: agx.GenericConstraintBodyData = constraintData.getBodyDataAt(index=1)

          # Similar/identical if one has access to the rigid body instances:
          # rb1_data: agx.GenericConstraintBodyData = constraintData.getBodyData(rb=rb1)
          # rb2_data: agx.GenericConstraintBodyData = constraintData.getBodyData(rb=rb2)

          rb1_data.setJacobian(0, agx.Vec3(), v1.cross(v2))
          rb2_data.setJacobian(0, agx.Vec3(), v2.cross(v1))

          constraintData.setViolation(0, v1 * v2)

    my_hinge.addElementaryConstraint("D1_VN", MyDot1(agx.Attachment.V, agx.Attachment.N))
    my_hinge.addElementaryConstraint("D1_UN", MyDot1(agx.Attachment.U, agx.Attachment.N))
  */
  class AGXPHYSICS_EXPORT GenericElementaryConstraint : public ElementaryConstraint
  {
    public:
      /**
      Construct given the number of rows in this elementary constraint.
      \note Using this constructor the enable flag is set to true.
      \param numRows - number of rows in this elementary constraint
      */
      GenericElementaryConstraint( UInt numRows );

      /**
      Construct given the number of rows in this elementary constraint and default enable flag.
      \param numRows - number of rows in this elementary constraint
      \param enable - true to be enabled by default, false disabled
      */
      GenericElementaryConstraint( UInt numRows, Bool enable );

      /**
      \return the attachment pair in the constraint
      */
      const AttachmentPair* getAttachmentPair() const;

      /**
      \return the cached data for this elementary constraint
      */
      GenericConstraintData* getConstraintData() const;

      /**
      \internal

      Enable (or disable) non-linear callback if this is an elementary
      contact constraint where the normal is at row 0 and the tangents
      occupies the rest of the rows.

      Note that this has to be called each time the contact point or
      bodies has changed.
      
      \param enable - true to enable, false to disable (disabled by default)
      \param primaryFrictionCoefficient - primary friction coefficient
      \param secondaryFrictionCoefficient - secondary friction coefficient
      \param contactPointPosition - position of the contact point in world frame
      */
      void setEnableMultiTangentNlCallback( Bool enable,
                                            Real primaryFrictionCoefficient,
                                            Real secondaryFrictionCoefficient,
                                            const Vec3& contactPointPosition );

      /**
      \internal

      \return true if enabled, otherwise false
      */
      Bool getEnableMultiTangentNlCallback() const;

    public:
      /**
      Callback when the elementary constraints in the constraint receives the
      prepare call. The return value defines if this elementary constraint is
      active or not (regardless of the enable flag). When this elementary
      constraint isn't active, onWriteData won't be called.
      \return true if this elementary constraint is active, false if inactive
      */
      virtual Bool onPrepare();

      /**
      Callback when to write Jacobian and constraint data if this elementary
      constraint is active.
      \param constraintData - constraint data for this elementary constraint
      */
      virtual void onWriteData( GenericConstraintData* constraintData );

#ifndef SWIG
    public:
      /**
      Performing callbacks and collecting data in serial.
      */
      void fetchData( const ConstraintRigidBodyContainer& bodies );
#endif

    public:
      virtual Bool isActive() const override;
#ifndef SWIG
      ConstraintNlmcpCallback* getNlCallback() const override;
#endif

      void setAttachmentPair( const AttachmentPair* ap ) override;

    protected:
      virtual ~GenericElementaryConstraint();

    protected:
      const AttachmentPair* m_attachmentPair;
      Bool m_active;
      GenericConstraintData m_constraintData;
      ref_ptr<Referenced> m_multiTangentNlCallback;
  };
}
